/**
 * hermes antispam proxy
 * Copyright (C) 2006, 2007 Juan José Gutiérrez de Quevedo <juanjo@gutierrezdequevedo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @author Juan José Gutiérrez de Quevedo <juanjo@gutierrezdequevedo.com>
 */
#include <string>
#include <memory>
#include <stdexcept>
#include <windows.h>

namespace {
    // Service configuration constants
    constexpr const char* SERVICE_NAME = "hermes anti-spam proxy";
    constexpr const char* SERVICE_SHORT_NAME = "hermes";
    constexpr const char* SERVICE_DESCRIPTION_TEXT = 
        "An anti-spam proxy using a combination of techniques like greylisting, dnsbl/dnswl, SPF, etc.";

    // Global state
    SERVICE_STATUS service_status;
    SERVICE_STATUS_HANDLE service_status_handle;
    extern bool quit;

    // Function declarations
    static void WINAPI service_main(DWORD argc, LPTSTR* argv);
    static void WINAPI handler(DWORD code);
    static int service_install();
    static int service_uninstall();
    extern int hermes_main(int argc, char** argv);

    // Helper class for managing command line parameters
    class Parameters {
    public:
        Parameters() {
            params_ = std::make_unique<char*[]>(2);
            params_[0] = new char[1];
            params_[0][0] = '\0';
            params_[1] = new char[1024]();
        }

        ~Parameters() {
            delete[] params_[0];
            delete[] params_[1];
        }

        char** get() { return params_.get(); }
        char* operator[](size_t index) { return params_[index]; }

    private:
        std::unique_ptr<char*[]> params_;
    };

    // Helper functions
    void show_error(const char* message) {
        MessageBox(NULL, message, ("Error from " SERVICE_NAME), MB_OK | MB_ICONERROR);
    }

    void show_message(const char* message) {
        MessageBox(NULL, message, ("Message from " SERVICE_NAME), MB_OK | MB_ICONINFORMATION);
    }

    void update_service_status(DWORD new_state) {
        service_status.dwCurrentState = new_state;
        SetServiceStatus(service_status_handle, &service_status);
    }
}

int WINAPI WinMain(HINSTANCE instance, HINSTANCE previous_instance, LPSTR cmdline, int cmdshow) {
    try {
        if (strncmp(cmdline, "-service", strlen("-service")) == 0) {
            SERVICE_TABLE_ENTRY service_table[] = {
                {const_cast<LPSTR>(SERVICE_SHORT_NAME), service_main},
                {NULL, NULL}
            };

            if (!StartServiceCtrlDispatcher(service_table)) {
                throw std::runtime_error("Error starting service dispatcher");
            }
        }
        else if (strncmp(cmdline, "-install", strlen("-install")) == 0) {
            return service_install();
        }
        else if (strncmp(cmdline, "-uninstall", strlen("-uninstall")) == 0) {
            return service_uninstall();
        }
        else {
            Parameters params;
            strncpy(params[1], cmdline, 1023);
            params[1][1023] = '\0';  // Ensure null termination
            return hermes_main(2, params.get());
        }
    }
    catch (const std::exception& e) {
        show_error(e.what());
        return -1;
    }

    return 0;
}

static int service_install() {
    char filename[1024] = {0};
    if (GetModuleFileName(NULL, filename, sizeof(filename)) == 0) {
        throw std::runtime_error("Error getting the file name of the process");
    }

    SC_HANDLE scm = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
    if (!scm) {
        throw std::runtime_error("Error opening connection to the Service Manager");
    }

    std::string exepath = "\"" + std::string(filename) + "\" -service";

    SC_HANDLE service_handle = CreateService(
        scm,                           // SCM handle
        SERVICE_SHORT_NAME,            // Service name
        SERVICE_NAME,                  // Display name
        SERVICE_ALL_ACCESS,            // Desired access
        SERVICE_WIN32_OWN_PROCESS,     // Service type
        SERVICE_AUTO_START,            // Start type
        SERVICE_ERROR_NORMAL,          // Error control
        exepath.c_str(),              // Executable path
        NULL, NULL, NULL, NULL, NULL   // Other parameters
    );

    if (!service_handle) {
        CloseServiceHandle(scm);
        throw std::runtime_error("Error creating service. Already installed?");
    }

    // Set service description
    SERVICE_DESCRIPTION service_description = {const_cast<LPSTR>(SERVICE_DESCRIPTION_TEXT)};
    ChangeServiceConfig2(service_handle, SERVICE_CONFIG_DESCRIPTION, &service_description);

    CloseServiceHandle(service_handle);
    CloseServiceHandle(scm);
    show_message("Service successfully installed");
    return 0;
}

static int service_uninstall() {
    SC_HANDLE scm = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
    if (!scm) {
        throw std::runtime_error("Error opening connection to the Service Manager");
    }

    SC_HANDLE service_handle = OpenService(scm, SERVICE_SHORT_NAME, SERVICE_QUERY_STATUS | DELETE);
    if (!service_handle) {
        CloseServiceHandle(scm);
        throw std::runtime_error("Error opening service");
    }

    SERVICE_STATUS status;
    if (!QueryServiceStatus(service_handle, &status)) {
        CloseServiceHandle(service_handle);
        CloseServiceHandle(scm);
        throw std::runtime_error("Error querying service");
    }

    if (status.dwCurrentState != SERVICE_STOPPED) {
        CloseServiceHandle(service_handle);
        CloseServiceHandle(scm);
        throw std::runtime_error(SERVICE_NAME " is still running. Stop it before trying to uninstall it");
    }

    if (!DeleteService(service_handle)) {
        CloseServiceHandle(service_handle);
        CloseServiceHandle(scm);
        throw std::runtime_error("Error deleting service");
    }

    CloseServiceHandle(service_handle);
    CloseServiceHandle(scm);
    show_message("Service successfully uninstalled");
    return 0;
}

static void WINAPI service_main(DWORD argc, LPTSTR* argv) {
    service_status = {
        .dwServiceType = SERVICE_WIN32,
        .dwControlsAccepted = SERVICE_ACCEPT_STOP,
        .dwWin32ExitCode = NO_ERROR,
        .dwServiceSpecificExitCode = NO_ERROR,
        .dwCheckPoint = 0,
        .dwWaitHint = 0
    };

    service_status_handle = RegisterServiceCtrlHandler(SERVICE_SHORT_NAME, handler);
    if (!service_status_handle) {
        return;
    }

    update_service_status(SERVICE_RUNNING);

    try {
        Parameters params;
        if (GetModuleFileName(NULL, params[1], 1024) == 0) {
            throw std::runtime_error("Error getting module filename");
        }

        char* config_path = strrchr(params[1], '\\');
        if (!config_path) {
            throw std::runtime_error("Error finding default config file");
        }

        *(++config_path) = '\0';
        strncat(params[1], "hermes.ini", strlen("hermes.ini"));

        hermes_main(2, params.get());

        update_service_status(SERVICE_STOP_PENDING);
        update_service_status(SERVICE_STOPPED);
    }
    catch (const std::exception& e) {
        show_error(e.what());
        update_service_status(SERVICE_STOPPED);
    }
}

static void WINAPI handler(DWORD code) {
    if (code == SERVICE_CONTROL_STOP) {
        quit = true;
        update_service_status(SERVICE_STOP_PENDING);
    }
}
